!MNH_LIC Copyright 2010-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!#########################
MODULE MODI_CPHASE_PROFILE
!#########################
!
INTERFACE
!
      SUBROUTINE CPHASE_PROFILE (PZHAT,PCPHASE,PCPHASE_PBL,PCPHASE_PROFILE,PTKEM)
! 
REAL, DIMENSION(:)              ,  INTENT(IN)  :: PZHAT       ! height level without orography
REAL                            ,  INTENT(IN)  :: PCPHASE     ! prescribed phase velocity
REAL                            ,  INTENT(IN)  :: PCPHASE_PBL ! prescribed phase velocity
REAL, DIMENSION(:,:)            ,  INTENT(OUT) :: PCPHASE_PROFILE ! profile of Cphase speed
REAL, DIMENSION(:,:),OPTIONAL   ,  INTENT(IN)  :: PTKEM ! TKE at t-dt
!
END SUBROUTINE CPHASE_PROFILE
!
END INTERFACE
!
END MODULE MODI_CPHASE_PROFILE
!
!     ##########################################################################
      SUBROUTINE CPHASE_PROFILE (PZHAT,PCPHASE,PCPHASE_PBL,PCPHASE_PROFILE,PTKEM)
!     ##########################################################################
!
!!****  *CPHASE_PROFILE* - defines a non-constant vertical profile for Cphase
!!                         velocity
!!
!!    PURPOSE
!!    -------
!
!!**  METHOD
!!    ------
!!   
!!        
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!	V. Masson & C. Lac     * Meteo France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    08/2010
!!      Escobar     9/11/2010 : array bound problem if NO Turb =>  PTKEM optional 
!!      C.Lac       06/2013   : correction and introduction of PCPHASE_PBL
!!      
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_CTURB
USE MODD_PARAMETERS

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,   ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
!
IMPLICIT NONE
!
!
!*       0.1   declarations of arguments
! 
! 
! 
REAL, DIMENSION(:)              ,  INTENT(IN)  :: PZHAT       ! height level without orography
REAL                            ,  INTENT(IN)  :: PCPHASE     ! prescribed phase velocity
REAL                            ,  INTENT(IN)  :: PCPHASE_PBL ! prescribed phase velocity
REAL, DIMENSION(:,:)            ,  INTENT(OUT) :: PCPHASE_PROFILE ! profile of Cphase speed
REAL, DIMENSION(:,:),OPTIONAL   ,  INTENT(IN)  :: PTKEM ! TKE at t-dt
!
!*       0.2   declarations of local variables
!
INTEGER             :: IKB       ! indice K Beginning in z direction 
INTEGER             :: IKE       ! indice K End       in z direction 
! 
#ifndef MNH_OPENACC
REAL, DIMENSION(SIZE(PCPHASE_PROFILE,1)) :: ZTKE, ZTKEMIN           
#else
REAL, DIMENSION(:), POINTER, CONTIGUOUS :: ZTKE, ZTKEMIN
#endif
INTEGER                                  :: JL,JK,JKTKE
INTEGER                                  :: JLMAX ! Intermediate variable to work around problem with CCE 15.0.1
!
!-------------------------------------------------------------------------------
!
!*       1.    PROLOGUE
!              --------
!
!*       1.1  Compute dimensions of arrays and other indices
! 
IKB = 1 + JPVEXT
IKE = SIZE(PCPHASE_PROFILE,2) - JPVEXT
JLMAX = SIZE(PCPHASE_PROFILE,1)
!
#ifdef MNH_OPENACC
CALL MNH_MEM_POSITION_PIN( 'CPHASE_PROFILE' )

CALL MNH_MEM_GET( ZTKE,    SIZE( PCPHASE_PROFILE, 1 ) )
CALL MNH_MEM_GET( ZTKEMIN, SIZE( PCPHASE_PROFILE, 1 ) )

!$acc data present( PZHAT, PCPHASE_PROFILE,ZTKE, ZTKEMIN )
#endif
!
!*       1.2  Initializations
!
!
!$acc kernels
PCPHASE_PROFILE(:,:) = 0.0
ZTKEMIN(:)           = PZHAT(IKE)
ZTKE(:)              = PZHAT(IKE-1)
!$acc end kernels
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!
IF (PRESENT(PTKEM)) THEN
!acc data present( PTKEM )
!$acc kernels
!$acc loop private( JKTKE )  independent
  DO JL = 1, JLMAX
    JKTKE=IKE-1
    DO JK = IKB, IKE-1
      IF (PTKEM(JL,JK) < 5.*XTKEMIN ) THEN
        ZTKE (JL) = PZHAT (JK)
        JKTKE = JK
        EXIT
      END IF
    END DO
    DO JK = JKTKE+1,IKE
      IF (PTKEM(JL,JK) == XTKEMIN ) THEN
        ZTKEMIN (JL) = PZHAT (JK)
        EXIT
      END IF
    END DO
  END DO
!$acc end kernels
!acc end data
ELSE
!$acc kernels
  ZTKE (:) = 1000.
  ZTKEMIN (:) = 2000.
!$acc end kernels
END IF
!
!$acc kernels
DO JL = 1, JLMAX
  DO JK = IKB, IKE
    IF (PZHAT(JK) > ZTKEMIN (JL) ) THEN
      PCPHASE_PROFILE(JL,JK) = PCPHASE
    ELSE IF  (PZHAT(JK) < ZTKE (JL) ) THEN
      PCPHASE_PROFILE(JL,JK) = PCPHASE_PBL
    ELSE
      PCPHASE_PROFILE(JL,JK) = 1./(ZTKEMIN (JL) - ZTKE (JL)) * &
                              ((PZHAT(JK) - ZTKE(JL)) * PCPHASE + (ZTKEMIN (JL) - PZHAT(JK)) * PCPHASE_PBL )
    END IF
  END DO
END DO
!$acc end kernels

#ifdef MNH_OPENACC
!$acc end data

!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'CPHASE_PROFILE' )
#endif
!
!-------------------------------------------------------------------------------
!
END SUBROUTINE CPHASE_PROFILE
