!MNH_LIC Copyright 1994-2021 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_NUCLEATION
INTERFACE
SUBROUTINE ICE4_NUCLEATION(ODSOFT, ODCOMPUTE, &
                           PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                           PRVT, &
                           PCIT, PRVHENI_MR, PB_TH, PB_RV, PB_RI)
IMPLICIT NONE
LOGICAL,              INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(:),INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(:),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(:),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(:),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(:),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_RV
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_RI
END SUBROUTINE ICE4_NUCLEATION
END INTERFACE
END MODULE MODI_ICE4_NUCLEATION
SUBROUTINE ICE4_NUCLEATION(ODSOFT, ODCOMPUTE, &
                           PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                           PRVT, &
                           PCIT, PRVHENI_MR, PB_TH, PB_RV, PB_RI)
!!
!!**  PURPOSE
!!    -------
!!      Computes the nucleation
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XGAMI,XGAMW,XMD,XMV,XTT,XEPSILO
USE MODD_PARAM_ICE,      ONLY: LFEEDBACKT
USE MODD_RAIN_ICE_PARAM, ONLY: XALPHA1,XALPHA2,XBETA1,XBETA2,XMNU0,XNU10,XNU20
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,              INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(:),INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(:),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(:),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(:),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(:),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_RV
REAL, DIMENSION(:),   INTENT(INOUT) :: PB_RI
!
!*       0.2  declaration of local variables
!
INTEGER                            :: JI
#ifndef MNH_OPENACC
LOGICAL, DIMENSION(:), allocatable :: GNEGT  ! Test where to compute the HEN process
LOGICAL, DIMENSION(:), allocatable :: GWORK, GWORK2
REAL, DIMENSION(:), allocatable    :: ZW ! work array
REAL, DIMENSION(:), allocatable    :: ZZW,      & ! Work array
                                      ZUSW,     & ! Undersaturation over water
                                      ZSSI        ! Supersaturation over ice
#else
LOGICAL, DIMENSION(:), pointer, contiguous :: GNEGT  ! Test where to compute the HEN process
LOGICAL, DIMENSION(:), pointer, contiguous :: GWORK, GWORK2
REAL,    DIMENSION(:), pointer, contiguous :: ZW ! work array
REAL,    DIMENSION(:), pointer, contiguous :: ZZW,      & ! Work array
                                              ZUSW,     & ! Undersaturation over water
                                              ZSSI        ! Supersaturation over ice
#endif

!$acc data present(ODCOMPUTE,PTHT,PPABST,PRHODREF,PEXN,PLSFACT,PT,PRVT,PCIT,PRVHENI_MR,PB_TH,PB_RV,PB_RI)

!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PTHT,"ICE4_NUCLEATION beg:PTHT")
  CALL MPPDB_CHECK(PPABST,"ICE4_NUCLEATION beg:PPABST")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_NUCLEATION beg:PRHODREF")
  CALL MPPDB_CHECK(PEXN,"ICE4_NUCLEATION beg:PEXN")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_NUCLEATION beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_NUCLEATION beg:PT")
  CALL MPPDB_CHECK(PRVT,"ICE4_NUCLEATION beg:PRVT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION beg:PCIT")
  CALL MPPDB_CHECK(PB_TH,"ICE4_NUCLEATION beg:PB_TH")
  CALL MPPDB_CHECK(PB_RV,"ICE4_NUCLEATION beg:PB_RV")
  CALL MPPDB_CHECK(PB_RI,"ICE4_NUCLEATION beg:PB_RI")
END IF

#ifndef MNH_OPENACC
allocate( gnegt (size( odcompute ) ) )
allocate( gwork (size( odcompute ) ) )
allocate( gwork2(size( odcompute ) ) )
allocate( zw    (size( odcompute ) ) )
allocate( zzw   (size( odcompute ) ) )
allocate( zusw  (size( odcompute ) ) )
allocate( zssi  (size( odcompute ) ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_NUCLEATION' )

CALL MNH_MEM_GET( gnegt,  size( odcompute ) )
CALL MNH_MEM_GET( gwork,  size( odcompute ) )
CALL MNH_MEM_GET( gwork2, size( odcompute ) )
CALL MNH_MEM_GET( zw,     size( odcompute ) )
CALL MNH_MEM_GET( zzw,    size( odcompute ) )
CALL MNH_MEM_GET( zusw,   size( odcompute ) )
CALL MNH_MEM_GET( zssi,   size( odcompute ) )

!$acc data present( GNEGT, GWORK, GWORK2, ZW, ZZW, ZUSW, ZSSI )
#endif

!$acc kernels
PRVHENI_MR(:)=0.
IF(.NOT. ODSOFT) THEN
  GNEGT(:)=PT(:)<XTT .AND. PRVT>XRTMIN(1) .AND. ODCOMPUTE(:)
  PRVHENI_MR(:)=0.
  ZSSI(:)=0.
  ZUSW(:)=0.
  ZZW(:)=0.
  !$mnh_do_concurrent( JI = 1 : size( odcompute ) )
    IF ( GNEGT(ji) ) THEN
#ifndef MNH_BITREP
      ZZW(JI)=ALOG(PT(JI))
      ZUSW(JI)=EXP(XALPW - XBETAW/PT(JI) - XGAMW*ZZW(JI))          ! es_w
      ZZW(JI)=EXP(XALPI - XBETAI/PT(JI) - XGAMI*ZZW(JI))           ! es_i
#else
      ZZW(JI)  = BR_LOG(PT(JI))
      ZUSW(JI) = BR_EXP(XALPW - XBETAW/PT(JI) - XGAMW*ZZW(JI))          ! es_w
      ZZW(JI)  = BR_EXP(XALPI - XBETAI/PT(JI) - XGAMI*ZZW(JI))          ! es_i
#endif
    END IF
  !$mnh_end_do()
  !$mnh_do_concurrent( JI = 1 : size( odcompute ) )
    IF ( GNEGT(ji) ) THEN
      ZZW(JI)=MIN(PPABST(JI)/2., ZZW(JI))             ! safety limitation
      ZSSI(JI)=PRVT(JI)*(PPABST(JI)-ZZW(JI)) / (XEPSILO*ZZW(JI)) - 1.0
                                                   ! Supersaturation over ice
      ZUSW(JI)=MIN(PPABST(JI)/2., ZUSW(JI))            ! safety limitation
      ZUSW(JI)=(ZUSW(JI)/ZZW(JI))*((PPABST(JI)-ZZW(JI))/(PPABST(JI)-ZUSW(JI))) - 1.0
                                 ! Supersaturation of saturated water vapor over ice
      !
      !*       3.1     compute the heterogeneous nucleation source RVHENI
      !
      !*       3.1.1   compute the cloud ice concentration
      !
      ZSSI(JI)=MIN(ZSSI(JI), ZUSW(JI)) ! limitation of SSi according to SSw=0
    END IF
  !$mnh_end_do()
  ZZW(:)=0.
  !$mnh_do_concurrent( JI = 1 : size( odcompute ) )
    GWORK(JI)  = GNEGT(JI) .AND. PT(JI)<XTT-5.0  .AND. ZSSI(JI)>0.0
    GWORK2(JI) = GNEGT(JI) .AND. PT(JI)<=XTT-2.0 .AND. PT(JI)>=XTT-5.0 .AND. ZSSI(JI)>0.0
  !$mnh_end_do()
#ifndef MNH_BITREP
#ifndef MNH_OPENACC
  WHERE(GWORK(:))
    ZZW(:)=XNU20*EXP(XALPHA2*ZSSI(:)-XBETA2)
  ELSEWHERE(GWORK2(:))
    ZZW(:)=MAX(XNU20*EXP(-XBETA2 ), &
               XNU10*EXP(-XBETA1*(PT(:)-XTT))*(ZSSI(:)/ZUSW(:))**XALPHA1)
  END WHERE
#else
!$acc loop independent
  DO JI=1,size( odcompute )
    IF (GWORK(JI)) THEN
      ZZW(JI)=XNU20*EXP(XALPHA2*ZSSI(JI)-XBETA2)
    ELSE IF (GWORK2(JI)) THEN
      ZZW(JI)=MAX(XNU20*(-XBETA2 ), &
              XNU10*EXP(-XBETA1*(PT(JI)-XTT))*(ZSSI(JI)/ZUSW(JI))**XALPHA1)
    END IF
  END DO
#endif
#else
#ifndef MNH_OPENACC
  WHERE(GWORK(:))
    ZZW(:)=XNU20*BR_EXP(XALPHA2*ZSSI(:)-XBETA2)
  ELSEWHERE(GWORK2(:))
    ZZW(:)=MAX(XNU20*BR_EXP(-XBETA2 ), &
               XNU10*BR_EXP(-XBETA1*(PT(:)-XTT))*BR_POW(ZSSI(:)/ZUSW(:),XALPHA1))
  END WHERE
#else
!$acc loop independent
  DO JI=1,size( odcompute )
    IF (GWORK(JI)) THEN
      ZZW(JI)=XNU20*BR_EXP(XALPHA2*ZSSI(JI)-XBETA2)
    ELSE IF (GWORK2(JI)) THEN
      ZZW(JI)=MAX(XNU20*BR_EXP(-XBETA2 ), &
              XNU10*BR_EXP(-XBETA1*(PT(JI)-XTT))*BR_POW(ZSSI(JI)/ZUSW(JI),XALPHA1))
    END IF
  END DO
#endif
#endif
  !$mnh_do_concurrent( JI = 1 : size( odcompute ) )
    ZZW(JI)=ZZW(JI)-PCIT(JI)
    ZZW(JI)=MIN(ZZW(JI), 50.E3) ! limitation provisoire a 50 l^-1
  !$mnh_end_do()
  WHERE(GNEGT(:))
    !
    !*       3.1.2   update the r_i and r_v mixing ratios
    !
    PRVHENI_MR(:)=MAX(ZZW(:), 0.0)*XMNU0/PRHODREF(:)
    PRVHENI_MR(:)=MIN(PRVT(:), PRVHENI_MR(:))
  END WHERE
  !Limitation due to 0 crossing of temperature
  IF(LFEEDBACKT) THEN
    !$mnh_do_concurrent( JI = 1 : size( odcompute ) )
      IF ( GNEGT(JI) ) THEN
        ZW(JI)=MIN(PRVHENI_MR(JI), &
                   MAX(0., (XTT/PEXN(JI)-PTHT(JI))/PLSFACT(JI))) / &
                   MAX(PRVHENI_MR(JI), 1.E-20)
      ELSE
        ZW(JI) = 0.
      END IF
    !$mnh_end_do()
  ELSE
    ZW(:)=1.
  ENDIF
  PRVHENI_MR(:)=PRVHENI_MR(:)*ZW(:)
  PCIT(:)=MAX(ZZW(:)*ZW(:)+PCIT(:), PCIT(:))
  !
  PB_RI(:)=PB_RI(:) + PRVHENI_MR(:)
  PB_RV(:)=PB_RV(:) - PRVHENI_MR(:)
  PB_TH(:)=PB_TH(:) + PRVHENI_MR(:)*PLSFACT(:)
ENDIF
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRVHENI_MR,"ICE4_NUCLEATION end:PRVHENI_MR")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION end:PCIT")
  CALL MPPDB_CHECK(PB_TH,"ICE4_NUCLEATION end:PB_TH")
  CALL MPPDB_CHECK(PB_RV,"ICE4_NUCLEATION end:PB_RV")
  CALL MPPDB_CHECK(PB_RI,"ICE4_NUCLEATION end:PB_RI")
END IF

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'ICE4_NUCLEATION' )
#endif

!$acc end data

END SUBROUTINE ICE4_NUCLEATION
